==> intel-oneapi-mkl-2022.1.0_223-3-aarch64.pkg.tar.xz: f_offload/dft/source/dp_real_3d_async.f90 <==
!===============================================================================
! Copyright 2021-2022 Intel Corporation.
!
! This software and the related documents are Intel copyrighted  materials,  and
! your use of  them is  governed by the  express license  under which  they were
! provided to you (License).  Unless the License provides otherwise, you may not
! use, modify, copy, publish, distribute,  disclose or transmit this software or
! the related documents without Intel's prior written permission.
!
! This software and the related documents  are provided as  is,  with no express
! or implied  warranties,  other  than those  that are  expressly stated  in the
! License.
!===============================================================================

! Content:
! A simple example of asynchronous double-precision real-to-complex, 
! complex-to-real in-place 3D FFT using Intel(R) oneAPI Math Kernel Library 
! (oneMKL) DFTI
!
!*****************************************************************************

include "mkl_dfti_omp_offload.f90"

program dp_real_3d_async
  use MKL_DFTI_OMP_OFFLOAD, forget => DFTI_DOUBLE, DFTI_DOUBLE => DFTI_DOUBLE_R
  use omp_lib, ONLY : omp_get_num_devices
  use, intrinsic :: ISO_C_BINDING

  ! Sizes of 2D transform
  integer, parameter :: N1 = 7
  integer, parameter :: N2 = 13
  integer, parameter :: N3 = 5
  
  integer, parameter :: halfN1plus1 = N1/2 + 1
  
  ! Arbitrary harmonic to test the FFT
  integer, parameter :: H1 = 1
  integer, parameter :: H2 = 2
  integer, parameter :: H3 = 3

  ! need double precision
  integer, parameter :: WP = selected_real_kind(15,307)

  ! Execution status
  integer :: status = 0, ignored_status

  real(WP), allocatable :: x (:)
  
  type(DFTI_DESCRIPTOR), POINTER :: hand

  hand => null()

  print *,"Example dp_real_3d_async"
  print *,"Asynchronous forward and backward double-precision real-to-complex",&
    &    " and complex-to-real in-place 3D transform"
  print *,"Configuration parameters:"
  print *,"DFTI_PRECISION      = DFTI_DOUBLE"
  print *,"DFTI_FORWARD_DOMAIN = DFTI_REAL"
  print *,"DFTI_DIMENSION      = 3"
  print '(" DFTI_LENGTHS        = /"I0","I0","I0"/" )', N1, N2, N3

  print *,"Create DFTI descriptor"
  status = DftiCreateDescriptor(hand, DFTI_DOUBLE, DFTI_REAL, 3, [N1,N2,N3])
  if (0 /= status) goto 999
  
  print *,"Set DFTI descriptor for CCE storage "
  status = DftiSetValue(hand, DFTI_CONJUGATE_EVEN_STORAGE, DFTI_COMPLEX_COMPLEX)
  if (0 /= status) goto 999
  
  print *,"Set input strides for DFTI descriptor (forward transform) "
  status = DftiSetValue(hand, DFTI_INPUT_STRIDES, &
                          [0, 1, 2*halfN1plus1, 2*halfN1plus1*N2])
  if (0 /= status) goto 999
  
  print *,"Set output strides for DFTI descriptor (forward transform) "
  status = DftiSetValue(hand, DFTI_OUTPUT_STRIDES, &
                          [0, 1, halfN1plus1, halfN1plus1*N2])
  if (0 /= status) goto 999

  print *,"Commit DFTI descriptor for forward FFT"
  !$omp target variant dispatch
  status = DftiCommitDescriptor(hand)
  !$omp end target variant dispatch
  if (0 /= status) goto 999

  print *,"Allocate array for input/output data"
  allocate ( x(2*halfN1plus1*N2*N3), STAT = status)
  if (0 /= status) goto 999

  print *,"Initialize input for forward transform"
  call init_r(x, N1, N2, N3, H1, H2, H3)

  print *,"Compute forward transform"
  !$omp target data map(tofrom:x)
  !$omp target variant dispatch use_device_ptr(x) nowait
  status = DftiComputeForward(hand, x)
  !$omp end target variant dispatch
  !$omp taskwait
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the result"
  status = verify_c(x, N1, N2, N3, H1, H2, H3)
  if (0 /= status) goto 999

  print *,"Initialize input for backward transform"
  call init_c(x, N1, N2, N3, H1, H2, H3)
  
  ! reset the strides of input and output data accordingly:
  print *,"Set input strides for DFTI descriptor (backward transform)"
  status = DftiSetValue(hand, DFTI_INPUT_STRIDES, &
                          [0, 1, halfN1plus1, halfN1plus1*N2])
  if (0 /= status) goto 999
  
  print *,"Set output strides for DFTI descriptor (backward transform) "
  status = DftiSetValue(hand, DFTI_OUTPUT_STRIDES, &
                          [0, 1, 2*halfN1plus1, 2*halfN1plus1*N2])
  if (0 /= status) goto 999
  
  ! we changed the configuration for the DFTI descriptor
  ! --> commit again
  print *,"Commit DFTI descriptor for backward transform "
  !$omp target variant dispatch
  status = DftiCommitDescriptor(hand)
  !$omp end target variant dispatch
  if (0 /= status) goto 999

  print *,"Compute backward transform"
  !$omp target data map(tofrom:x)
  !$omp target variant dispatch use_device_ptr(x) nowait
  status = DftiComputeBackward(hand, x)
  !$omp end target variant dispatch
  !$omp taskwait
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the result"
  status = verify_r(x, N1, N2, N3, H1, H2, H3)
  if (0 /= status) goto 999

100 continue

  print *,"Release the DFTI descriptor"
  ignored_status = DftiFreeDescriptor(hand)

  if (allocated(x)) then
      print *,"Deallocate input data array"
      deallocate(x)
  endif

  if (status == 0) then
    print *,"TEST PASSED"
    call exit(0)
  else
    print *,"TEST FAILED"
    call exit(1)
  endif

999 print '("  Error, status = ",I0)', status
  goto 100

contains

  ! Compute mod(K*L,M) accurately
  pure real(WP) function moda(k,l,m)
    integer, intent(in) :: k,l,m
    integer*8 :: k8
    k8 = k
    moda = real(mod(k8*l,m),WP)
  end function moda

  ! Initialize array x to produce unit peaks at transform(H1, H2)
  subroutine init_r(x, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: x(:)

    integer k1, k2, k3, halfN1plus1
    real(WP), parameter :: TWOPI = 6.2831853071795864769_WP
    real(WP) :: TWOPI_phase, factor
    if( mod(2*(N1-H1), N1) == 0 .AND. &
        mod(2*(N2-H2), N2) == 0 .AND. &
        mod(2*(N3-H3), N3) == 0) then
      factor = 1.0_WP
    else
      factor = 2.0_WP
    endif
    
    halfN1plus1 = N1/2 + 1
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, N1
          TWOPI_phase = TWOPI*&
          (moda(k1-1,H1,N1)/N1 + moda(k2-1,H2,N2)/N2 + moda(k3-1,H3,N3)/N3)
          x((k3-1)*2*halfN1plus1*N2 + (k2-1)*2*halfN1plus1 + k1) = &
              factor*cos(TWOPI_phase)/(N1*N2*N3)
        end do
      end do
    end do
  end subroutine init_r

  ! Verify that y is unit peak at (H1,H2,H3)
  integer function verify_c(y, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: y(:)

    integer k1, k2, k3, halfN1plus1, idx
    real(WP) err, errthr, maxerr
    complex(WP) :: res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 2.5 * log(real(N1*N2*N3,WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    halfN1plus1 = N1/2 + 1
    maxerr = 0.0_WP
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, halfN1plus1
          if  (   (mod(k1-1-H1, N1)==0 &
                    .AND. mod(k2-1-H2, N2)==0 &
                    .AND. mod(k3-1-H3, N3)==0) &
            .OR.  (mod(1-k1-H1, N1)==0 &
                    .AND. mod(1-k2-H2, N2)==0 &
                    .AND. mod(1-k3-H3, N3)==0)) then
            res_exp = 1.0_WP
          else
            res_exp = 0.0_WP
          end if
          idx = (k3-1)*2*halfN1plus1*N2 &
                + (k2-1)*2*halfN1plus1 + 2*k1 - 1
          res_got = DCMPLX(y(idx), y(idx + 1))
          err = abs(res_got - res_exp)
          maxerr = max(err,maxerr)
          if (.not.(err < errthr)) then
            print '("  y("I0","I0","I0"):"$)', k1, k2, k3
            print '(" expected ("G24.17","G24.17"),"$)', res_exp
            print '(" got ("G24.17","G24.17"),"$)', res_got
            print '(" err "G10.3)', err
            print *," Verification FAILED"
            verify_c = 100
            return
          end if
        end do
      end do
    end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify_c = 0
  end function verify_c
  
  ! Initialize complex array y to produce unit peaks at x(H1, H2)
  subroutine init_c(y, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: y(:)

    integer k1, k2, k3, halfN1plus1, idx
    real(WP), parameter :: TWOPI = 6.2831853071795864769_WP
    real(WP) :: TWOPI_phase
    
    halfN1plus1 = N1/2 + 1
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, halfN1plus1
          TWOPI_phase = TWOPI*&
            (moda(k1-1,H1,N1)/N1 + moda(k2-1,H2,N2)/N2 + moda(k3-1,H3,N3)/N3)
          idx = (k3-1)*2*halfN1plus1*N2 + (k2-1)*2*halfN1plus1 + 2*k1 - 1
          y(idx)      =  cos(TWOPI_phase)/(N1*N2*N3)
          y(idx + 1)  = -sin(TWOPI_phase)/(N1*N2*N3)
        end do
      end do
    end do
  end subroutine init_c
  
  ! Verify that x(k1, k2) is unit peak at k1 = H1 and k2 = H2
  integer function verify_r(x, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: x(:)

    integer k1, k2, k3, halfN1plus1
    real(WP) err, errthr, maxerr
    real(WP) res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 2.5 * log(real(N1*N2*N3, WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    halfN1plus1 = N1/2 +1 
    maxerr = 0.0_WP
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, N1
          if (mod(k1-1-H1,N1)==0 .AND. &
              mod(k2-1-H2,N2)==0 .AND. &
              mod(k3-1-H3,N3)==0) then
            res_exp = 1.0_WP
          else
            res_exp = 0.0_WP
          end if
          res_got = x((k3-1)*2*halfN1plus1*N2 + (k2-1)*2*halfN1plus1 + k1)
          err = abs(res_got - res_exp)
          maxerr = max(err,maxerr)
          if (.not.(err < errthr)) then
            print '("  x("I0","I0","I0"): "$)', k1, k2, k3
            print '(" expected "G24.17","$)', res_exp
            print '(" got "G24.17","$)', res_got
            print '(" err "G10.3)', err
            print *," Verification FAILED"
            verify_r = 100
            return
          end if
        end do
      end do
    end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify_r = 0
  end function verify_r

end program dp_real_3d_async

==> intel-oneapi-mkl-2023.2.0_49495-2-aarch64.pkg.tar.xz: f_offload/dft/source/dp_real_3d_async.f90 <==
!===============================================================================
! Copyright 2021-2022 Intel Corporation.
!
! This software and the related documents are Intel copyrighted  materials,  and
! your use of  them is  governed by the  express license  under which  they were
! provided to you (License).  Unless the License provides otherwise, you may not
! use, modify, copy, publish, distribute,  disclose or transmit this software or
! the related documents without Intel's prior written permission.
!
! This software and the related documents  are provided as  is,  with no express
! or implied  warranties,  other  than those  that are  expressly stated  in the
! License.
!===============================================================================

! Content:
! A simple example of asynchronous double-precision real-to-complex, 
! complex-to-real in-place 3D FFT using Intel(R) oneAPI Math Kernel Library 
! (oneMKL) DFTI
!
!*****************************************************************************

include "mkl_dfti_omp_offload.f90"

program dp_real_3d_async
  use MKL_DFTI_OMP_OFFLOAD, forget => DFTI_DOUBLE, DFTI_DOUBLE => DFTI_DOUBLE_R
  use omp_lib, ONLY : omp_get_num_devices
  use, intrinsic :: ISO_C_BINDING

  ! Sizes of 2D transform
  integer, parameter :: N1 = 7
  integer, parameter :: N2 = 13
  integer, parameter :: N3 = 5
  
  integer, parameter :: halfN1plus1 = N1/2 + 1
  
  ! Arbitrary harmonic to test the FFT
  integer, parameter :: H1 = 1
  integer, parameter :: H2 = 2
  integer, parameter :: H3 = 3

  ! need double precision
  integer, parameter :: WP = selected_real_kind(15,307)

  ! Execution status
  integer :: status = 0, ignored_status

  real(WP), allocatable :: x (:)
  
  type(DFTI_DESCRIPTOR), POINTER :: hand

  hand => null()

  print *,"Example dp_real_3d_async"
  print *,"Asynchronous forward and backward double-precision real-to-complex",&
    &    " and complex-to-real in-place 3D transform"
  print *,"Configuration parameters:"
  print *,"DFTI_PRECISION      = DFTI_DOUBLE"
  print *,"DFTI_FORWARD_DOMAIN = DFTI_REAL"
  print *,"DFTI_DIMENSION      = 3"
  print '(" DFTI_LENGTHS        = /"I0","I0","I0"/" )', N1, N2, N3

  print *,"Create DFTI descriptor"
  status = DftiCreateDescriptor(hand, DFTI_DOUBLE, DFTI_REAL, 3, [N1,N2,N3])
  if (0 /= status) goto 999
  
  print *,"Set DFTI descriptor for CCE storage "
  status = DftiSetValue(hand, DFTI_CONJUGATE_EVEN_STORAGE, DFTI_COMPLEX_COMPLEX)
  if (0 /= status) goto 999
  
  print *,"Set input strides for DFTI descriptor (forward transform) "
  status = DftiSetValue(hand, DFTI_INPUT_STRIDES, &
                          [0, 1, 2*halfN1plus1, 2*halfN1plus1*N2])
  if (0 /= status) goto 999
  
  print *,"Set output strides for DFTI descriptor (forward transform) "
  status = DftiSetValue(hand, DFTI_OUTPUT_STRIDES, &
                          [0, 1, halfN1plus1, halfN1plus1*N2])
  if (0 /= status) goto 999

  print *,"Commit DFTI descriptor for forward FFT"
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch
#else
  !$omp target variant dispatch
#endif
  status = DftiCommitDescriptor(hand)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  if (0 /= status) goto 999

  print *,"Allocate array for input/output data"
  allocate ( x(2*halfN1plus1*N2*N3), STAT = status)
  if (0 /= status) goto 999

  print *,"Initialize input for forward transform"
  call init_r(x, N1, N2, N3, H1, H2, H3)

  print *,"Compute forward transform"
  !$omp target data map(tofrom:x)
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch nowait
#else
  !$omp target variant dispatch use_device_ptr(x) nowait
#endif
  status = DftiComputeForward(hand, x)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  !$omp taskwait
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the result"
  status = verify_c(x, N1, N2, N3, H1, H2, H3)
  if (0 /= status) goto 999

  print *,"Initialize input for backward transform"
  call init_c(x, N1, N2, N3, H1, H2, H3)
  
  ! reset the strides of input and output data accordingly:
  print *,"Set input strides for DFTI descriptor (backward transform)"
  status = DftiSetValue(hand, DFTI_INPUT_STRIDES, &
                          [0, 1, halfN1plus1, halfN1plus1*N2])
  if (0 /= status) goto 999
  
  print *,"Set output strides for DFTI descriptor (backward transform) "
  status = DftiSetValue(hand, DFTI_OUTPUT_STRIDES, &
                          [0, 1, 2*halfN1plus1, 2*halfN1plus1*N2])
  if (0 /= status) goto 999
  
  ! we changed the configuration for the DFTI descriptor
  ! --> commit again
  print *,"Commit DFTI descriptor for backward transform "
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch
#else
  !$omp target variant dispatch
#endif
  status = DftiCommitDescriptor(hand)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  if (0 /= status) goto 999

  print *,"Compute backward transform"
  !$omp target data map(tofrom:x)
#if defined(ONEMKL_USE_OPENMP_VERSION) && (ONEMKL_USE_OPENMP_VERSION >= 202011)
  !$omp dispatch nowait
#else
  !$omp target variant dispatch use_device_ptr(x) nowait
#endif
  status = DftiComputeBackward(hand, x)
#if !defined(ONEMKL_USE_OPENMP_VERSION) || (ONEMKL_USE_OPENMP_VERSION < 202011)
  !$omp end target variant dispatch
#endif
  !$omp taskwait
  !$omp end target data
  if (0 /= status) goto 999

  print *,"Verify the result"
  status = verify_r(x, N1, N2, N3, H1, H2, H3)
  if (0 /= status) goto 999

100 continue

  print *,"Release the DFTI descriptor"
  ignored_status = DftiFreeDescriptor(hand)

  if (allocated(x)) then
      print *,"Deallocate input data array"
      deallocate(x)
  endif

  if (status == 0) then
    print *,"TEST PASSED"
    call exit(0)
  else
    print *,"TEST FAILED"
    call exit(1)
  endif

999 print '("  Error, status = ",I0)', status
  goto 100

contains

  ! Compute mod(K*L,M) accurately
  pure real(WP) function moda(k,l,m)
    integer, intent(in) :: k,l,m
    integer*8 :: k8
    k8 = k
    moda = real(mod(k8*l,m),WP)
  end function moda

  ! Initialize array x to produce unit peaks at transform(H1, H2)
  subroutine init_r(x, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: x(:)

    integer k1, k2, k3, halfN1plus1
    real(WP), parameter :: TWOPI = 6.2831853071795864769_WP
    real(WP) :: TWOPI_phase, factor
    if( mod(2*(N1-H1), N1) == 0 .AND. &
        mod(2*(N2-H2), N2) == 0 .AND. &
        mod(2*(N3-H3), N3) == 0) then
      factor = 1.0_WP
    else
      factor = 2.0_WP
    endif
    
    halfN1plus1 = N1/2 + 1
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, N1
          TWOPI_phase = TWOPI*&
          (moda(k1-1,H1,N1)/N1 + moda(k2-1,H2,N2)/N2 + moda(k3-1,H3,N3)/N3)
          x((k3-1)*2*halfN1plus1*N2 + (k2-1)*2*halfN1plus1 + k1) = &
              factor*cos(TWOPI_phase)/(N1*N2*N3)
        end do
      end do
    end do
  end subroutine init_r

  ! Verify that y is unit peak at (H1,H2,H3)
  integer function verify_c(y, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: y(:)

    integer k1, k2, k3, halfN1plus1, idx
    real(WP) err, errthr, maxerr
    complex(WP) :: res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 2.5 * log(real(N1*N2*N3,WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    halfN1plus1 = N1/2 + 1
    maxerr = 0.0_WP
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, halfN1plus1
          if  (   (mod(k1-1-H1, N1)==0 &
                    .AND. mod(k2-1-H2, N2)==0 &
                    .AND. mod(k3-1-H3, N3)==0) &
            .OR.  (mod(1-k1-H1, N1)==0 &
                    .AND. mod(1-k2-H2, N2)==0 &
                    .AND. mod(1-k3-H3, N3)==0)) then
            res_exp = 1.0_WP
          else
            res_exp = 0.0_WP
          end if
          idx = (k3-1)*2*halfN1plus1*N2 &
                + (k2-1)*2*halfN1plus1 + 2*k1 - 1
          res_got = DCMPLX(y(idx), y(idx + 1))
          err = abs(res_got - res_exp)
          maxerr = max(err,maxerr)
          if (.not.(err < errthr)) then
            print '("  y("I0","I0","I0"):"$)', k1, k2, k3
            print '(" expected ("G24.17","G24.17"),"$)', res_exp
            print '(" got ("G24.17","G24.17"),"$)', res_got
            print '(" err "G10.3)', err
            print *," Verification FAILED"
            verify_c = 100
            return
          end if
        end do
      end do
    end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify_c = 0
  end function verify_c
  
  ! Initialize complex array y to produce unit peaks at x(H1, H2)
  subroutine init_c(y, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: y(:)

    integer k1, k2, k3, halfN1plus1, idx
    real(WP), parameter :: TWOPI = 6.2831853071795864769_WP
    real(WP) :: TWOPI_phase
    
    halfN1plus1 = N1/2 + 1
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, halfN1plus1
          TWOPI_phase = TWOPI*&
            (moda(k1-1,H1,N1)/N1 + moda(k2-1,H2,N2)/N2 + moda(k3-1,H3,N3)/N3)
          idx = (k3-1)*2*halfN1plus1*N2 + (k2-1)*2*halfN1plus1 + 2*k1 - 1
          y(idx)      =  cos(TWOPI_phase)/(N1*N2*N3)
          y(idx + 1)  = -sin(TWOPI_phase)/(N1*N2*N3)
        end do
      end do
    end do
  end subroutine init_c
  
  ! Verify that x(k1, k2) is unit peak at k1 = H1 and k2 = H2
  integer function verify_r(x, N1, N2, N3, H1, H2, H3)
    integer N1, N2, N3, H1, H2, H3
    real(WP) :: x(:)

    integer k1, k2, k3, halfN1plus1
    real(WP) err, errthr, maxerr
    real(WP) res_exp, res_got

    ! Note, this simple error bound doesn't take into account error of
    ! input data
    errthr = 2.5 * log(real(N1*N2*N3, WP)) / log(2.0_WP) * EPSILON(1.0_WP)
    print '("  Check if err is below errthr " G10.3)', errthr

    halfN1plus1 = N1/2 +1 
    maxerr = 0.0_WP
    do k3 = 1, N3
      do k2 = 1, N2
        do k1 = 1, N1
          if (mod(k1-1-H1,N1)==0 .AND. &
              mod(k2-1-H2,N2)==0 .AND. &
              mod(k3-1-H3,N3)==0) then
            res_exp = 1.0_WP
          else
            res_exp = 0.0_WP
          end if
          res_got = x((k3-1)*2*halfN1plus1*N2 + (k2-1)*2*halfN1plus1 + k1)
          err = abs(res_got - res_exp)
          maxerr = max(err,maxerr)
          if (.not.(err < errthr)) then
            print '("  x("I0","I0","I0"): "$)', k1, k2, k3
            print '(" expected "G24.17","$)', res_exp
            print '(" got "G24.17","$)', res_got
            print '(" err "G10.3)', err
            print *," Verification FAILED"
            verify_r = 100
            return
          end if
        end do
      end do
    end do
    print '("  Verified,  maximum error was " G10.3)', maxerr
    verify_r = 0
  end function verify_r

end program dp_real_3d_async
